# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-01 02:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0030_add_data_duplicate"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="duplicated",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]
